@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 # flake8: noqa
 from u115.api import (API, Passport, RequestHandler, Request, Response,
+                      RequestsLWPCookieJar, RequestsMozillaCookieJar,
                       Torrent, Task, TorrentFile, File, Directory,
                       APIError, TaskError, AuthenticationError,
-                      InvalidAPIAccess, RequestFailure)
+                      InvalidAPIAccess, RequestFailure, JobError)
