@@ -9,12 +9,32 @@ import requests
 import time
 from hashlib import sha1
 from bs4 import BeautifulSoup
+from requests.cookies import RequestsCookieJar
 from u115 import conf
 from u115.utils import (get_timestamp, get_utcdatetime, string_to_datetime,
-                        eval_path, quote, utf8_encode, txt_type, PY3, download)
+                        eval_path, quote, unquote, utf8_encode, txt_type, PY3)
+from homura import download
+
+if PY3:
+    from http import cookiejar as cookielib
+else:
+    import cookielib
 
 USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_4) \
 AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.94 Safari/537.36'
+LOGIN_URL = 'http://passport.115.com/?ct=login&ac=ajax&is_ssl=1'
+LOGOUT_URL = 'http://passport.115.com/?ac=logout'
+CHECKPOINT_URL = 'http://passport.115.com/?ct=ajax&ac=ajax_check_point'
+
+
+class RequestsLWPCookieJar(cookielib.LWPCookieJar, RequestsCookieJar):
+    """RequestsCookieJar compatible LWPCookieJar"""
+    pass
+
+
+class RequestsMozillaCookieJar(cookielib.MozillaCookieJar, RequestsCookieJar):
+    """RequestsCookieJar compatible RequestsMozillaCookieJar"""
+    pass
 
 
 class RequestHandler(object):
@@ -127,26 +147,91 @@ class API(object):
 
     num_tasks_per_page = 30
     web_api_url = 'http://web.api.115.com/files'
+    aps_natsort_url = 'http://aps.115.com/natsort/files.php'
 
-    def __init__(self, auto_logout=True):
+    def __init__(self, auto_logout=True, persistent=False,
+                 cookies_filename=None, cookies_type='LWPCookieJar'):
         """
         :param bool auto_logout: whether to logout automatically when
             :class:`.API` object is destroyed
+        :param bool persistent: whether to use persistent session that stores
+            cookies on disk
+        :param str cookies_filename: path to the cookies file, use default
+            path (`~/.115cookies`) if none
+        :param str cookies_type: a string representing
+            :class:`cookielib.FileCookieJar` subclass,
+            `LWPCookieJar` (default) or `MozillaCookieJar`
         """
         self.auto_logout = auto_logout
+        self.persistent = persistent
+        self.cookies_filename = cookies_filename
+        self.cookies_type = cookies_type
         self.passport = None
         self.http = RequestHandler()
+        # Cached variabled to decrease API hits
+        self._user_id = None
+        self._username = None
         self._signatures = {}
         self._upload_url = None
         self._lixian_timestamp = None
+        self._root_directory = None
+        self._downloads_directory = None
+        self._torrents_directory = None
+        self._task_count = None
+        self._task_quota = None
+        if self.persistent:
+            self.load_cookies()
+
+    def __del__(self):
+        if self.auto_logout and self.has_logged_in:
+            self.logout()
+        if not self.auto_logout and self.has_logged_in:
+            if self.persistent:
+                self.save_cookies()
+
+    def _reset_cache(self):
+        self._user_id = None
+        self._username = None
+        self._signatures = {}
+        self._upload_url = None
+        self._lixian_timestamp = None
+        self._root_directory = None
         self._downloads_directory = None
         self._torrents_directory = None
         self._task_count = None
         self._task_quota = None
 
-    def __del__(self):
-        if self.auto_logout and self.has_logged_in:
-            self.logout()
+    def _init_cookies(self):
+        # RequestsLWPCookieJar or RequestsMozillaCookieJar
+        cookies_class = globals()['Requests' + self.cookies_type]
+        f = self.cookies_filename or conf.COOKIES_FILENAME
+        self.cookies = cookies_class(f)
+
+    def load_cookies(self, ignore_discard=True, ignore_expires=True):
+        """Load cookies from the file"""
+        self._init_cookies()
+        if os.path.exists(self.cookies.filename):
+            self.cookies.load(ignore_discard=ignore_discard,
+                              ignore_expires=ignore_expires)
+            self._reset_cache()
+
+    def save_cookies(self, ignore_discard=True, ignore_expires=True):
+        """Save cookies to the file"""
+        if not isinstance(self.cookies, cookielib.FileCookieJar):
+            m = 'Cookies must be a cookielib.FileCookieJar object to be saved.'
+            raise APIError(m)
+        self.cookies.save(ignore_discard=ignore_discard,
+                          ignore_expires=ignore_expires)
+
+    @property
+    def cookies(self):
+        """Cookies getter shortcut"""
+        return self.http.session.cookies
+
+    @cookies.setter
+    def cookies(self, cookies):
+        """Cookies setter shortcut"""
+        self.http.session.cookies = cookies
 
     def login(self, username=None, password=None,
               section='default'):
@@ -168,14 +253,13 @@ class API(object):
             password = credential['password']
 
         passport = Passport(username, password)
-        r = self.http.post(passport.login_url, passport.form)
+        r = self.http.post(LOGIN_URL, passport.form)
 
         if r.state is True:
             # Bind this passport to API
             self.passport = passport
             passport.data = r.content['data']
-            passport.user_id = r.content['data']['USER_ID']
-            passport.status = 'LOGGED_IN'
+            self._user_id = r.content['data']['USER_ID']
             return True
         else:
             msg = None
@@ -184,37 +268,64 @@ class API(object):
                     msg = 'Account does not exist.'
                 elif r.content['err_name'] == 'passwd':
                     msg = 'Password is incorrect.'
-            passport.status = 'FAILED'
             raise AuthenticationError(msg)
 
+    def get_user_info(self):
+        return self._req_get_user_aq()
+
+    @property
+    def user_id(self):
+        if self._user_id is None:
+            if self.has_logged_in:
+                self._user_id = self._req_get_user_aq()['data']['uid']
+            else:
+                raise AuthenticationError('Not logged in.')
+        return self._user_id
+
+    @property
+    def username(self):
+        if self._username is None:
+            if self.has_logged_in:
+                self._username = self._get_username()
+            else:
+                raise AuthenticationError('Not logged in.')
+        return self._username
+
     @property
     def has_logged_in(self):
         """Check whether the API has logged in"""
-        if self.passport is not None and self.passport.user_id is not None:
-            params = {'user_id': self.passport.user_id}
-            r = self.http.get(self.passport.checkpoint_url, params=params)
-            if r.state is False:
-                return True
+        r = self.http.get(CHECKPOINT_URL)
+        if r.state is False:
+            return True
+        # If logged out, flush cache
+        self._reset_cache()
         return False
 
     def logout(self):
         """Log out"""
-        self.http.get(self.passport.logout_url)
-        self.passport.status = 'LOGGED_OUT'
+        self.http.get(LOGOUT_URL)
+        self._reset_cache()
         return True
 
+    @property
+    def root_directory(self):
+        """Root directory"""
+        if self._root_directory is None:
+            self._load_root_directory()
+        return self._root_directory
+
     @property
     def downloads_directory(self):
         """Default directory for downloaded files"""
         if self._downloads_directory is None:
-            self._load_lixian_space()
+            self._load_downloads_directory()
         return self._downloads_directory
 
     @property
     def torrents_directory(self):
         """Default directory that stores uploaded torrents"""
         if self._torrents_directory is None:
-            self._load_lixian_space()
+            self._load_torrents_directory()
         return self._torrents_directory
 
     @property
@@ -263,16 +374,12 @@ class API(object):
             return t
         return t.submit()
 
-    def add_task_url(self):
+    def add_task_url(self, target_url):
         """
-        Added a new URL task (VIP only)
-
-        .. warning::
-
-           This method has not yet implemented.
+        :param str target_url: the URL of the file that to be downloaded
 
         """
-        raise NotImplementedError
+        return self._req_lixian_add_task_url(target_url)
 
     def get_storage_info(self, human=False):
         """
@@ -331,6 +438,27 @@ class API(object):
                  show_progress=show_progress, resume=resume,
                  auto_retry=auto_retry)
 
+    def search(self, keyword, count=30):
+        """
+        Search files or directories
+
+        :param str keyword: keyword
+        :param int count: number of entries to be listed
+        """
+        kwargs = {}
+        kwargs['search_value'] = keyword
+        root = self.root_directory
+        entries = root._load_entries(func=self._req_files_search,
+                                     count=count, page=1, **kwargs)
+
+        res = []
+        for entry in entries:
+            if 'pid' in entry:
+                res.append(_instantiate_directory(self, entry))
+            else:
+                res.append(_instantiate_file(self, entry))
+        return res
+
     def _req_offline_space(self):
         """Required before accessing lixian tasks"""
         url = 'http://115.com/'
@@ -354,7 +482,7 @@ class API(object):
         self._load_signatures()
         data = {
             'page': page,
-            'uid': self.passport.user_id,
+            'uid': self.user_id,
             'sign': self._signatures['offline_space'],
             'time': self._lixian_timestamp,
         }
@@ -370,7 +498,7 @@ class API(object):
 
     def _req_lixian_get_id(self, torrent=False):
         """Get `cid` of lixian space directory"""
-        url = 'http://115.com/lixian/'
+        url = 'http://115.com/'
         params = {
             'ct': 'lixian',
             'ac': 'get_id',
@@ -395,7 +523,7 @@ class API(object):
         data = {
             'pickcode': u.pickcode,
             'sha1': u.sha,
-            'uid': self.passport.user_id,
+            'uid': self.user_id,
             'sign': self._signatures['offline_space'],
             'time': self._lixian_timestamp,
         }
@@ -420,7 +548,26 @@ class API(object):
             'info_hash': t.info_hash,
             'wanted': wanted,
             'savepath': t.name,
-            'uid': self.passport.user_id,
+            'uid': self.user_id,
+            'sign': self._signatures['offline_space'],
+            'time': self._lixian_timestamp,
+        }
+        req = Request(method='POST', url=url, params=params, data=data)
+        res = self.http.send(req)
+        if res.state:
+            return True
+        else:
+            print(res.content.get('error_msg'))
+            raise RequestFailure('Failed to create new task.')
+
+    def _req_lixian_add_task_url(self, target_url):
+
+        self._load_signatures()
+        url = 'http://115.com/lixian/'
+        params = {'ct': 'lixian', 'ac': 'add_task_url'}
+        data = {
+            'url': target_url,
+            'uid': self.user_id,
             'sign': self._signatures['offline_space'],
             'time': self._lixian_timestamp,
         }
@@ -439,7 +586,7 @@ class API(object):
         params = {'ct': 'lixian', 'ac': 'task_del'}
         data = {
             'hash[0]': t.info_hash,
-            'uid': self.passport.user_id,
+            'uid': self.user_id,
             'sign': self._signatures['offline_space'],
             'time': self._lixian_timestamp,
         }
@@ -450,9 +597,44 @@ class API(object):
         else:
             raise RequestFailure('Failed to delete the task.')
 
-    def _req_files(self, cid, offset, limit, o='user_ptime', asc=0, aid=1,
-                   show_dir=1, code=None, scid=None, snap=0, natsort=None,
-                   source=None):
+    def _req_aps_natsort_files(self, cid, offset, limit, o='file_name',
+                               asc=1, aid=1, show_dir=1, code=None, scid=None,
+                               snap=0, natsort=1, source=None, type=0,
+                               format='json', star=None, is_share=None):
+        """
+        When :func:`API._req_files` is called with `o='filename'` and
+            `natsort=1`, API access will fail
+            and :func:`API._req_aps_natsort_files` is subsequently called with
+            the same kwargs. Refer to the implementation in
+            :func:`Directory.list`
+        """
+        params = locals()
+        del params['self']
+        req = Request(method='GET', url=self.aps_natsort_url, params=params)
+        res = self.http.send(req)
+        if res.state:
+            return res.content
+        else:
+            raise RequestFailure('Failed to access files API.')
+
+    def _req_files(self, cid, offset, limit, o='user_ptime', asc=1, aid=1,
+                   show_dir=1, code=None, scid=None, snap=0, natsort=1,
+                   source=None, type=0, format='json', star=None,
+                   is_share=None):
+        """
+        :param int type: type of files to be displayed
+
+            * '' (empty string): marked
+            * None: all
+            * 0: all
+            * 1: documents
+            * 2: images
+            * 3: music
+            * 4: video
+            * 5: zipped
+            * 6: applications
+            * 99: files only
+        """
         params = locals()
         del params['self']
         req = Request(method='GET', url=self.web_api_url, params=params)
@@ -462,6 +644,65 @@ class API(object):
         else:
             raise RequestFailure('Failed to access files API.')
 
+    def _req_files_search(self, offset, limit, search_value, aid=-1,
+                          date=None, pick_code=None, source=None, type=0,
+                          format='json'):
+        params = locals()
+        del params['self']
+        url = self.web_api_url + '/search'
+        req = Request(method='GET', url=url, params=params)
+        res = self.http.send(req)
+        if res.state:
+            return res.content
+        else:
+            raise RequestFailure('Failed to access files API.')
+
+    def _req_files_edit(self, fid, file_name=None, is_mark=0):
+        """Edit a file or directory"""
+        url = self.web_api_url + '/edit'
+        data = locals()
+        del data['self']
+        req = Request(method='POST', url=url, data=data)
+        res = self.http.send(req)
+        if res.state:
+            return res.content
+        else:
+            raise RequestFailure('Failed to access files API.')
+
+    def _req_files_add(self, pid, cname):
+        """
+        Add a directory
+        :param int pid: parent directory id
+        :param str cname: directory name
+        """
+        url = self.web_api_url + '/add'
+        data = locals()
+        del data['self']
+        req = Request(method='POST', url=url, data=data)
+        res = self.http.send(req)
+        if res.state:
+            return res.content
+        else:
+            raise RequestFailure('Failed to access files API.')
+
+    def _req_files_move(self, pid, fids):
+        """
+        Move files or directories
+        :param int pid: target directory id
+        :param list fids: a list of ids of files or directories to be moved
+        """
+        url = self.web_api_url + '/move'
+        data = {}
+        data['pid'] = pid
+        for i, fid in enumerate(fids):
+            data['fid[%d]' % i] = fid
+        req = Request(method='POST', url=url, data=data)
+        res = self.http.send(req)
+        if res.state:
+            return res.content
+        else:
+            raise RequestFailure('Failed to access files API.')
+
     def _req_file(self, file_id):
         url = self.web_api_url + '/file'
         data = {'file_id': file_id}
@@ -547,9 +788,23 @@ class API(object):
             return True
         else:
             msg = 'Failed to delete this file or directory.'
+            if 'errno' in res.content:
+                if res.content['errno'] == 990005:
+                    raise JobError()
             print(res.content['error'])
             raise APIError(msg)
 
+    def _req_get_user_aq(self):
+        url = 'http://my.115.com/'
+        data = {
+            'ct': 'ajax',
+            'ac': 'get_user_aq'
+        }
+        req = Request(method='POST', url=url, data=data)
+        res = self.http.send(req)
+        if res.state:
+            return res.content
+
     def _load_signatures(self, force=True):
         if not self._signatures or force:
             self._req_offline_space()
@@ -574,13 +829,32 @@ class API(object):
         if str(kwargs['pid']) != str(cid):
             return Directory(api=self, **kwargs)
 
-    def _load_lixian_space(self):
-        """Load downloads and torrents directory"""
+    def _load_root_directory(self):
+        """
+        Load root directory, which has a cid of 0
+        """
+        kwargs = self._req_directory(0)
+        self._root_directory = Directory(api=self, **kwargs)
+
+    def _load_torrents_directory(self):
+        """
+        Load torrents directory
+
+        If it does not exist yet, this request will cause the system to create
+        one
+        """
+        r = self._req_lixian_get_id(torrent=True)
+        self._downloads_directory = self._load_directory(r['cid'])
+
+    def _load_downloads_directory(self):
+        """
+        Load downloads directory
+
+        If it does not exist yet, this request will cause the system to create
+        one
+        """
         r = self._req_lixian_get_id(torrent=False)
-        downloads_cid = r['dest_cid']
-        torrent_cid = r['cid']
-        self._downloads_directory = self._load_directory(downloads_cid)
-        self._torrents_directory = self._load_directory(torrent_cid)
+        self._downloads_directory = self._load_directory(r['cid'])
 
     def _load_upload_url(self):
         res = self._parse_src_js_var('upload_config_h5')
@@ -601,10 +875,13 @@ class API(object):
         pattern = "%s = (.*);" % (variable.upper())
         m = re.search(pattern, text)
         if not m:
-            msg = 'Cannot parse source JavaScript for %s' % variable
+            msg = 'Cannot parse source JavaScript for %s.' % variable
             raise APIError(msg)
         return json.loads(m.group(1).strip())
 
+    def _get_username(self):
+        return unquote(self.cookies.get('OOFL'))
+
 
 class Base(object):
     def __repr__(self):
@@ -633,25 +910,14 @@ class Passport(Base):
     :ivar dict form: a dictionary of POST data to login
     :ivar int user_id: user ID of the authenticated user
     :ivar dict data: data returned upon login
-    :ivar str status: status of this passport
-
-        * `NEW`: passport is newly created
-        * `LOGGED_IN`: successfully logged in with this passport
-        * `LOGGED_OUT`: logged out
-        * `FAILED`: failed to log in
 
     """
-    login_url = 'http://passport.115.com/?ct=login&ac=ajax&is_ssl=1'
-    logout_url = 'http://passport.115.com/?ac=logout'
-    checkpoint_url = 'http://passport.115.com/?ct=ajax&ac=ajax_check_point'
 
     def __init__(self, username, password):
         self.username = username
         self.password = password
         self.form = self._form()
-        self.user_id = None
         self.data = None
-        self.status = 'NEW'
 
     def _form(self):
         vcode = self._vcode()
@@ -728,7 +994,7 @@ class BaseFile(Base):
                 self._deleted = True
                 return True
         else:
-            return APIError('This file or directory is already deleted.')
+            raise APIError('This file or directory is already deleted.')
 
     @property
     def is_deleted(self):
@@ -824,7 +1090,7 @@ class Directory(BaseFile):
     max_entries_per_load = 24  # Smaller than 24 may cause abnormal result
 
     def __init__(self, api, cid, name, pid, count=-1,
-                 date_created=None, pickcode=None,
+                 date_created=None, pickcode=None, is_root=False,
                  *args, **kwargs):
         super(Directory, self).__init__(api, cid, name)
 
@@ -835,6 +1101,11 @@ class Directory(BaseFile):
         self.pickcode = pickcode
         self._parent = None
 
+    @property
+    def is_root(self):
+        """Whether this directory is the root directory"""
+        return int(self.cid) == 0
+
     @property
     def parent(self):
         """Parent directory that holds this directory"""
@@ -864,54 +1135,75 @@ class Directory(BaseFile):
         self.name = r['name']
         self._count = r['count']
 
-    def _load_entries(self, count, page=1, order='user_ptime',
-                      asc=0, show_dir=1, entries=None):
+    def _load_entries(self, func, count, page=1, entries=None, **kwargs):
         """
         Load entries
 
+        :param function func: function (:func:`API._req_files` or
+            :func:`API._req_search`) that returns entries
         :param int count: number of entries to load. This value should never
             be greater than self.count
-        :param int page: page number (starting from 0)
+        :param int page: page number (starting from 1)
 
         """
         if entries is None:
             entries = []
+        res = \
+            func(offset=(page - 1) * self.max_entries_per_load,
+                 limit=self.max_entries_per_load,
+                 **kwargs)
         loaded_entries = [
-            entry for entry in
-            self.api._req_files(cid=self.cid,
-                                offset=(page - 1) * self.max_entries_per_load,
-                                limit=self.max_entries_per_load,
-                                o=order, show_dir=show_dir,
-                                asc=asc)['data'][:count]
+            entry for entry in res['data'][:count]
         ]
+        total_count = res['count']
+        # count should never be greater than total_count
+        if count > total_count:
+            count = total_count
         if count <= self.max_entries_per_load:
             return entries + loaded_entries
         else:
             cur_count = count - self.max_entries_per_load
             return self._load_entries(
-                count=cur_count, page=page + 1, order=order, asc=asc,
-                show_dir=show_dir, entries=entries + loaded_entries)
+                func=func, count=cur_count, page=page + 1,
+                entries=entries + loaded_entries, **kwargs)
 
-    def list(self, count=30, order='user_ptime', asc=False, show_dir=True):
+    def list(self, count=30, order='user_ptime', asc=False, show_dir=True,
+             natsort=True):
         """
         List directory contents
 
         :param int count: number of entries to be listed
-        :param str order: originally named `o`
+        :param str order: order of entries, originally named `o`. This value
+            may be one of `user_ptime` (default), `file_size` and `file_name`
         :param bool asc: whether in ascending order
         :param bool show_dir: whether to show directories
+        :param bool natsort: whether to use natural sort
 
         Return a list of :class:`.File` or :class:`.Directory` objects
         """
         if self.cid is None:
             return False
-        asc = 1 if asc is True else 0
-        show_dir = 1 if show_dir else 0
+        kwargs = {}
+        # `cid` is the only required argument
+        kwargs['cid'] = self.cid
+        kwargs['asc'] = 1 if asc is True else 0
+        kwargs['show_dir'] = 1 if show_dir is True else 0
+        kwargs['natsort'] = 1 if natsort is True else 0
+        kwargs['o'] = order
         if self.count <= count:
             # count should never be greater than self.count
             count = self.count
-        entries = self._load_entries(count, page=1, order=order, asc=asc,
-                                     show_dir=show_dir)
+        try:
+            entries = self._load_entries(func=self.api._req_files,
+                                         count=count, page=1, **kwargs)
+        # When natsort=1 and order='file_name', API access will fail
+        except RequestFailure as e:
+            if natsort is True and order == 'file_name':
+                entries = \
+                    self._load_entries(func=self.api._req_aps_natsort_files,
+                                       count=count, page=1, **kwargs)
+            else:
+                raise e
         res = []
         for entry in entries:
             if 'pid' in entry:
@@ -979,13 +1271,13 @@ class Task(Directory):
         Delete task (does not influence its corresponding directory)
 
         :return: whether deletion is successful
-        :raise: :class:`.APIError` if the task is already deleted
+        :raise: :class:`.TaskError` if the task is already deleted
         """
         if not self._deleted:
             if self.api._req_lixian_task_del(self):
                 self._deleted = True
                 return True
-        return APIError('This task is already deleted.')
+        raise TaskError('This task is already deleted.')
 
     @property
     def is_deleted(self):
@@ -1248,3 +1540,14 @@ class InvalidAPIAccess(APIError):
 class RequestFailure(APIError):
     """Request failure"""
     pass
+
+
+class JobError(APIError):
+    """Job running error (request multiple similar jobs simultaneously)"""
+    def __init__(self, *args, **kwargs):
+        content = kwargs.pop('content', None)
+        self.content = content
+        if not args:
+            msg = 'Your account has a similar job running. Try again later.'
+            args = (msg,)
+        super(JobError, self).__init__(*args, **kwargs)
